/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <math.h>
#include <log.h>
#include "command_internal.h"

#define Min3(x,y,z) ((x < y) ? ((z < x) ? z : x) : ((z < y) ? z : y))

static char * type = "Transient Analysis";

struct _commandPrivate {
	double tstep;
	double tstop;
	double tstart;
	double tmax;
	int uic;
};

#define MIN_STEP (p->tstep * 1e-8) /* from spice 3f5 */

static int execute(command_ *r, state_ *state)
{
	commandPrivate_ *p;
	double tN = 0, tP = 0, hN = 0, hP = 0, hBrk = 0, hLTE; 
	double nextBreak, lastBreak = 0;
	int brk = 0;
	deviceNIType niType;
	
	ReturnErrIf(r == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	ReturnErrIf(stateSetTStep(state, p->tstep));
	ReturnErrIf(stateSetTStop(state, p->tstop));
	if(!stateMinbreakIsSet(state)) {
		ReturnErrIf(stateSetMinbreak(state, p->tstep * 5e-5));
	}
	
	ReturnErrIf(deviceDBLoad(r->devices, state));
	ReturnErrIf(commandSolve(r, state, stateGetITL1(state)) < 0);
	
	/* Loosly Based on the Algortithum in "The SPICE Book", page 311 */
	niType = NI_BE;
	hN = ((p->tstep < p->tmax) ? p->tstep : p->tmax)/100;
	tN = 0;
	tP = 0;
	while(tN < p->tstop) {
		
		ReturnErrIf(matrixRecord(r->matrix, tN, brk));
		nextBreak = deviceDBCalcNextBreak(r->devices, state, lastBreak, tN);
		ReturnErrIf(isnan(nextBreak));
		
		if(!brk) {
			if(nextBreak < (tN + hN)) {
				lastBreak = nextBreak;
				brk = 1;
				hP = hN;
				hBrk = nextBreak - tN;
				hN = hBrk;
			}
		} else {
			brk = 0;
			hN = Min3(0.1*hBrk, hN, 0.1*hP);
			niType = NI_BE;
		}
		
		while(1) {
			tN = tP + hN;
			ReturnErrIf(deviceDBStep(r->devices, state, tN, niType));
			if(commandReSolve(r, state, stateGetITL4(state)) >= 0) {
				hLTE = deviceDBCalcNextStep(r->devices, state, niType);
				ReturnErrIf(isnan(hLTE));
				if(hLTE < (0.9 * hN)) {
					hN = hLTE;
					ReturnErrIf(hN < MIN_STEP, "TIMESTEP TOO SMALL (LTE)");
				} else {
					break;
				}
			} else {
				Warn("failed to linearize");
				hN = hN/8;
				ReturnErrIf(hN < MIN_STEP, "TIMESTEP TOO SMALL (LIN)");
				niType = NI_BE;
			}
			ReturnErrIf(matrixRecall(r->matrix));
		}
		
		hN = Min3(hLTE, 2*hN, p->tmax);
		if(tP != 0.0) {
			niType = NI_TRAP;
		}
		tP = tN;
	}
	
	return 0;
}

int commandConfigTransient(command_ *r, double tstep, double tstop, 
		double tstart, double tmax, int uic)
{
	commandPrivate_ *p;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->private != NULL);
	
	Debug("Configuring %s %p", type, r);
	
	r->type = type;
	r->execute = execute;
	
	r->private =  calloc(1, sizeof(commandPrivate_));
	ReturnErrIf(r->private == NULL);
	p = r->private;
	
	p->tstep = tstep;
	p->tstop = tstop;
	if(isnan(tstart))
		p->tstart = 0;
	else
		p->tstart = tstart;
	if(isnan(tmax))
		p->tmax = (p->tstep < ((p->tstop-p->tstart)/50)) ? p->tstep : 
				((p->tstop-p->tstart)/50);
	else
		p->tmax = tmax;
	p->uic = uic;
	
	return 0;
}
