/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <log.h>
#include "command_internal.h"

/*****************************************************************************/

static char *defaultType = "None";

static int linearize(command_ *r, state_ *state, int itl)
{
	int count = 0, linear;
	while(1) {
		linear = deviceDBLinearize(r->devices, state);
		ReturnErrIf(linear == LIN_ERR);
		if(linear == LIN_LINEAR)
			break;
		ReturnErrIf(matrixReSolve(r->matrix));
		ReturnErrIf(++count > itl);
	};
	return count;
}
	
int commandSolve(command_ *r, state_ *state, int itl)
{
	int count;
	ReturnErrIf(r == NULL);
	ReturnErrIf(state == NULL);
	ReturnErrIf(itl < 1);
	
	ReturnErrIf(matrixSolve(r->matrix));
	
	count = linearize(r, state, itl);
	ReturnErrIf(count < 0);
	
	return count;
}

int commandReSolve(command_ *r, state_ *state, int itl)
{
	int count;
	ReturnErrIf(r == NULL);
	ReturnErrIf(state == NULL);
	ReturnErrIf(itl < 1);
	
	ReturnErrIf(matrixReSolve(r->matrix));
	
	count = linearize(r, state, itl);
	ReturnErrIf(count < 0);
	
	return count;
}

int commandDestroy(command_ **r)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying Command %p", (*r));
	
	if((*r)->private != NULL)
		free((*r)->private);
	
	free(*r);
	*r = NULL;
	return 0;
}

command_ * commandNew(command_ *r,  matrix_ *matrix, deviceDB_ *devices)
{
	ReturnNULLIf(r != NULL);
	r = calloc(1, sizeof(command_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Command %p", r);
	r->type = defaultType;
	r->matrix = matrix;
	r->devices = devices;
	
	return r;
}

/*****************************************************************************/

struct _commandDB {
	command_ *command;
	commandDB_ *next;
};

int commandDBExecute(commandDB_ *r, state_ *state)
{
	commandDB_ *ptr;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(state == NULL);
	
	for(ptr = r->next; ptr != NULL; ptr = ptr->next) {
		ReturnErrIf(ptr->command == NULL);
		ReturnErrIf(ptr->command->execute == NULL);
		ReturnErrIf(ptr->command->execute(ptr->command, state));
	}
	
	return 0;
}

int commandDBAttach(commandDB_ *r, command_ *command)
{
	commandDB_ *newElement = NULL;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(command == NULL);
	
	newElement = calloc(1, sizeof(commandDB_));
	ReturnErrIf(newElement == NULL);
	newElement->command = command;
	newElement->next = r->next;
	r->next = newElement;
	
	return 0;
}

int commandDBDestroy(commandDB_ **r)
{
	commandDB_ *ptr, *ptrNext;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying Command DB %p", (*r));
	
	for(ptr = (*r); ptr != NULL; ptr = ptrNext) {
		ptrNext = ptr->next;
		if(ptr->command != NULL) {
			if(ptr->command->unconfig != NULL) {
				ReturnErrIf(ptr->command->unconfig(ptr->command));
			}
			ReturnErrIf(commandDestroy(&ptr->command));
		}
		free(ptr);
	}
	
	return 0;
}

commandDB_ * commandDBNew(commandDB_ *r)
{
	ReturnNULLIf(r != NULL);
	r = calloc(1, sizeof(commandDB_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Command DB %p", r);
	
	return r;
}

