/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <time.h>
#include <ctype.h>

#include <log.h>
#include "deck_internal.h"

static char * readFile(char *filename)
{
	char *buffer;
	long int length;
	FILE *inFile = NULL;
	
	if(!strcmp(filename, "stdin")) {
		inFile = stdin;
	} else {
		inFile = fopen(filename, "rb");
		ReturnNULLIf(inFile == NULL, "Failed to open include file %s", filename);
	}
	
	/* determine file length */
	ReturnNULLIf(fseek(inFile, 0L, SEEK_END), "Couldn't find end of file");
	length = ftell(inFile);
	ReturnNULLIf(length < 0, "Failed to find length of file");
	ReturnNULLIf(fseek(inFile, 0L, SEEK_SET), "Failed to reset file pointer");
	
	/* allocate memory and copy file into local buffer */
	buffer = (char *)malloc(sizeof(char) * (length+1));
	ReturnNULLIf(buffer == NULL);
	Debug("Copying %li bytes from file", length);
	ReturnNULLIf(fread(buffer, 1, (size_t) (length), inFile) < 0,
			"Failed to read from file");
	buffer[length-1] = 0x0; /* Make sure it's null terminated */
	
	fclose(inFile);
	
	return buffer;
}

/*****************************************************************************/

struct _includeDB {
	char *filename;
	char *buffer;
	includeDB_ *next;
};

static int includeDBDestroy(includeDB_ **r)
{
	includeDB_ *ptr, *ptrNext;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying Name DB %p", (*r));
	
	for(ptr = (*r); ptr != NULL; ptr = ptrNext) {
		ptrNext = ptr->next;
		if(ptr->buffer != NULL) {
			free(ptr->buffer);
		}
		free(ptr);
	}
	
	return 0;
}

static includeDB_ * includeDBNew(includeDB_ *r)
{
	ReturnNULLIf(r != NULL);
	r = calloc(1, sizeof(includeDB_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Include DB %p", r);
	
	return r;
}

char * includeDBAddFile(includeDB_ *r, char *filename)
{
	includeDB_ *ptr;
	
	ReturnNULLIf(r == NULL);
	ReturnNULLIf(filename == NULL);
	
	for(ptr = r; ptr != NULL; ptr = ptr->next) {
		if(ptr->filename != NULL) {
			ReturnNULLIf(!strcmp(filename, ptr->filename), "Recursive include");
		}
	}
	
	ptr = includeDBNew(ptr);
	ReturnNULLIf(ptr == NULL);
	ptr->buffer = readFile(filename);
	ReturnNULLIf(ptr->buffer == NULL);
	ptr->filename = filename;
	ptr->next = r->next;
	r->next = ptr;
	
	return ptr->buffer;
}

/*****************************************************************************/

int deckWriteRaw(deck_ *r, FILE *outFile)
{
	time_t currentTime;
	struct tm localTime;
	char timeString[26];
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(outFile == NULL);
	
	fprintf(outFile, "Title: %s\n", r->name);
	/* Get time string */
	ReturnErrIf(time(&currentTime) == -1);
	ReturnErrIf(localtime_r(&currentTime, &localTime) == NULL);
	ReturnErrIf(asctime_r(&localTime, timeString) == NULL);
	fprintf(outFile, "Date: %s", timeString);
	fprintf(outFile, "Plotname: Transient Analysis\n");
	fprintf(outFile, "Flags: real\n");
	
	ReturnErrIf(matrixWriteRaw(r->matrix, outFile));
	
	return 0;
}

int deckProcess(deck_ *r)
{
	ReturnErrIf(r == NULL);
	/* execute commands */
	ReturnErrIf(commandDBExecute(r->commands, r->state));
	return 0;
}

int deckDestroy(deck_ **r)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying deck %p", *r);
	
	matrixDestroy(&(*r)->matrix);
	stateDestroy(&((*r)->state));
	commandDBDestroy(&(*r)->commands);
	deviceDBDestroy(&(*r)->devices);
	subcktDBDestroy(&(*r)->subckts);
	includeDBDestroy(&(*r)->includes);
	
	if((*r)->buffer != NULL)
		free((*r)->buffer);
	
	free(*r);
	*r = NULL;
	return 0;
}

static int init(deck_ *r, char *filename)
{
	Debug("Creating deck %p", r);
	
	r->buffer = readFile(filename);
	ReturnErrIf(r->buffer == NULL);
	r->name = r->buffer;
	r->filename = filename;
	
	r->devices = deviceDBNew(r->devices);
	ReturnErrIf(r->devices == NULL);
	r->subckts = subcktDBNew(r->subckts);
	ReturnErrIf(r->subckts == NULL);
	r->includes = includeDBNew(r->includes);
	ReturnErrIf(r->includes == NULL);
	r->commands = commandDBNew(r->commands);
	ReturnErrIf(r->commands == NULL);
	r->state = stateNew(r->state);
	ReturnErrIf(r->state == NULL);
	r->matrix = matrixNew(r->matrix);
	ReturnErrIf(r->matrix == NULL);	
	
	ReturnErrIf(deckParse(r));
	
	ReturnErrIf(matrixInit(r->matrix, r->state));
	
	return 0;
}

deck_ * deckNew(deck_ *r, char *filename)
{
	ReturnNULLIf(r != NULL);
	ReturnNULLIf(filename == NULL);
	r = calloc(1, sizeof(deck_));
	ReturnNULLIf(r == NULL, "Malloc Failed");
	
	if(init(r, filename)) {
		deckDestroy(&r);
	}
	return r;
}

