/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <math.h>
#include <log.h>
#include "device_internal.h"

/*****************************************************************************/

static char *defaultType = "None";

int deviceDestroy(device_ **r)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying Device %p", (*r));
	
	if((*r)->unconfig != NULL) {
		ReturnErrIf((*r)->unconfig(*r));
	}
	
	if((*r)->private != NULL)
		free((*r)->private);
	
	free(*r);
	*r = NULL;
	return 0;
}

device_ * deviceNew(device_ *r,  matrix_ *matrix, char *refdes, row_ *pin[], 
		int numPins)
{
	int i;
	ReturnNULLIf(r != NULL);
	ReturnNULLIf(pin == NULL);
	ReturnNULLIf(numPins < 1);
	ReturnNULLIf(numPins > DEVICE_MAX_PINS);
	
	r = calloc(1, sizeof(device_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Device %s %p", refdes, r);
	
	r->matrix = matrix;
	r->refdes = refdes;
	for(i = 0; i < numPins; i++) {
		r->pin[i] = pin[i];
	}
	r->numPins = numPins;
	
	r->type = defaultType;
	
	return r;
}

/*****************************************************************************/

struct _deviceDB {
	device_ *device;
	deviceDB_ *next;
};

int deviceDBLoad(deviceDB_ *r, state_ *state)
{
	deviceDB_ *ptr;
	ReturnErrIf(r == NULL);
	
	for(ptr = r->next; ptr != NULL; ptr = ptr->next) {
		ReturnErrIf(ptr->device == NULL);
		if(ptr->device->load != NULL) {
			ReturnErrIf(ptr->device->load(ptr->device, state));
		}
	}
	
	return 0;
}

deviceLinear deviceDBLinearize(deviceDB_ *r, state_ *state)
{
	deviceDB_ *ptr;
	int linear = LIN_LINEAR, allLinear = LIN_LINEAR;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(state == NULL);
	
	for(ptr = r->next; ptr != NULL; ptr = ptr->next) {
		ReturnErrIf(ptr->device == NULL);
		if(ptr->device->linearize != NULL) {
			linear = ptr->device->linearize(ptr->device, state);
			ReturnErrIf(linear == LIN_ERR);
			if(linear == LIN_NOT_LINEAR)
				allLinear = LIN_NOT_LINEAR;	
		}
	}
	
	return linear;
}

int deviceDBStep(deviceDB_ *r, state_ *state, double time, deviceNIType niType)
{
	deviceDB_ *ptr;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(state == NULL);
	
	for(ptr = r->next; ptr != NULL; ptr = ptr->next) {
		ReturnErrIf(ptr->device == NULL);
		if(ptr->device->step != NULL) {
			ReturnErrIf(ptr->device->step(ptr->device, state, time, niType));
		}
	}
	
	return 0;
}

double deviceDBCalcNextStep(deviceDB_ *r, state_ *state, deviceNIType niType)
{
	deviceDB_ *ptr;
	double nextStep, nextStepMin;
	
	ReturnNaNIf(r == NULL);
	ReturnNaNIf(state == NULL);
	
	nextStepMin = stateGetTStop(state);
	ReturnNaNIf(isnan(nextStepMin));
	for(ptr = r->next; ptr != NULL; ptr = ptr->next) {
		ReturnNaNIf(ptr->device == NULL);
		if(ptr->device->nextStep != NULL) {
			nextStep = ptr->device->nextStep(ptr->device, state, niType);
			ReturnNaNIf(isnan(nextStep));
			nextStepMin = Min2(nextStep, nextStepMin);
		}
	}
	
	return nextStepMin;
}

double deviceDBCalcNextBreak(deviceDB_ *r, state_ *state, double lastBreak, 
		double time)
{
	deviceDB_ *ptr;
	double nextBreak, nextBreakMin;
	
	ReturnNaNIf(r == NULL);
	ReturnNaNIf(state == NULL);
	
	nextBreakMin = stateGetTStop(state);
	ReturnNaNIf(isnan(nextBreakMin));
	for(ptr = r->next; ptr != NULL; ptr = ptr->next) {
		ReturnNaNIf(ptr->device == NULL);
		if(ptr->device->nextBreak != NULL) {
			nextBreak = ptr->device->nextBreak(ptr->device, state, time);
			if(nextBreak > (lastBreak + stateGetMinbreak(state))) {
				nextBreakMin = Min2(nextBreakMin, nextBreak);
			}
		}
	}
	
	return nextBreakMin;
}

int deviceDBAttach(deviceDB_ *r, device_ *device)
{
	deviceDB_ *newElement;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(device == NULL);
	
	newElement = calloc(1, sizeof(deviceDB_));
	ReturnErrIf(newElement == NULL);
	newElement->device = device;
	newElement->next = r->next;
	r->next = newElement;
	
	return 0;
}

int deviceDBDestroy(deviceDB_ **r)
{
	deviceDB_ *ptr, *ptrNext;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying Device DB %p", (*r));
	
	for(ptr = (*r); ptr != NULL; ptr = ptrNext) {
		ptrNext = ptr->next;
		if(ptr->device != NULL) {
			ReturnErrIf(deviceDestroy(&ptr->device));
		}
		free(ptr);
	}
	
	return 0;
}

deviceDB_ * deviceDBNew(deviceDB_ *r)
{
	ReturnNULLIf(r != NULL);
	r = calloc(1, sizeof(deviceDB_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Device DB %p", r);
	
	return r;
}

