/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <log.h>
#include "matrix_internal.h"


#include "slu_ddefs.h"

static char *name = "SuperLU";

struct _matrixLibrary {
	SuperMatrix A;
	SuperMatrix B;
	SuperMatrix X;
	SuperMatrix L;
	SuperMatrix U;
	int *perm_r;
    int *perm_c;
	int *etree;
	char equed[1];
	int info;
	double *R;
	double *C;
	double rpg;
	double rcond;
	double *ferr;
	double *berr;
	mem_usage_t mem_usage;
	superlu_options_t options;
    SuperLUStat_t stat;
};

static int solve(matrix_ *r)
{
	matrixLibrary_ *p;
	
	ReturnErrIf(r == NULL);
	p = r->library;
	ReturnErrIf(p == NULL);
	
	/* 
	 memcpy(r->X, r->B, r->lenXB*sizeof(double));
	 dgssv(&p->options, &p->A, p->perm_c, p->perm_r, &p->L, &p->U, &p->B, 
			&p->stat, &info);
	*/
	
	dgssvx(&p->options, &p->A, p->perm_c, p->perm_r, p->etree, p->equed, 
			p->R, p->C, &p->L, &p->U, NULL, 0, &p->B, &p->X, &p->rpg, 
			&p->rcond, p->ferr, p->berr, &p->mem_usage, &p->stat, &p->info);
	
	ReturnErrIf(p->info < 0, "SuperLU: %ith arg had illegal value", -p->info);
	if(p->info > 0) {
		ReturnErrIf((p->info <= p->A.ncol), "SuperLU: U is singular");
		ReturnErrIf((p->info == (p->A.ncol + 1)), 
				"SuperLU: RCOND is singular to wokring precision");
		ReturnErrIf((p->info > (p->A.ncol + 1)), 
				"SuperLU: memory allocation error");
	}
	
	return 0;
}

static int reSolve(matrix_ *r)
{
	matrixLibrary_ *p;
	
	ReturnErrIf(r == NULL);
	p = r->library;
	ReturnErrIf(p == NULL);
	
	p->options.Fact = SamePattern_SameRowPerm;
	if(solve(r)) {
		p->options.Fact = DOFACT;
		ReturnErr("SuperLU Failed");
	}
	
	return 0;
}

static int unconfig(matrix_ *r)
{
	matrixLibrary_ *p;
	
	ReturnErrIf(r == NULL);
	p = r->library;
	ReturnErrIf(p == NULL);
	
	Debug("Unconfiguring Solution %p", r);
	
	SUPERLU_FREE(p->perm_r);
	SUPERLU_FREE(p->perm_c);
	SUPERLU_FREE(p->etree);
	SUPERLU_FREE(p->R);
    SUPERLU_FREE(p->C);
	SUPERLU_FREE(p->ferr);
    SUPERLU_FREE(p->berr);
	Destroy_CompCol_Matrix(&p->A);
	r->A = NULL;
	r->aRow = NULL;
	r->aColStart = NULL;
	Destroy_SuperMatrix_Store(&p->B);
	r->B = NULL;
	Destroy_SuperMatrix_Store(&p->X);
	r->X = NULL;
	StatFree(&p->stat);
	return 0;
}

int matrixInitSuperLU(matrix_ *r)
{
	matrixLibrary_ *p;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->library != NULL);
	
	Debug("Configuring Solution %p", r);
	
	r->name = name;
	r->unconfig = unconfig;
	r->solve = solve;
	r->reSolve = reSolve;
	
	r->library = calloc(1, sizeof(matrixLibrary_));
	ReturnErrIf(r->library == NULL);
	
	p = r->library;
	
	/* Create solution A in the format expected by SuperLU. */
	dCreate_CompCol_Matrix(&p->A, r->lenXB, r->lenXB, r->lenA, r->A,
			r->aRow, r->aColStart, SLU_NC, SLU_D, SLU_GE);
	
	/* Create right-hand solution matrix X. */
	dCreate_Dense_Matrix(&p->B, r->lenXB, 1, r->B, r->lenXB,
			SLU_DN, SLU_D, SLU_GE);
	
	dCreate_Dense_Matrix(&p->X, r->lenXB, 1, r->X, r->lenXB,
			SLU_DN, SLU_D, SLU_GE);
	
	/* Create permitation matrices. */
	p->perm_r = intMalloc(r->lenXB);
	ReturnErrIf(p->perm_r == NULL);
	p->perm_c = intMalloc(r->lenXB);
	ReturnErrIf(p->perm_r == NULL);
	p->etree = intMalloc(r->lenXB);
	ReturnErrIf(p->etree == NULL);
	p->R = (double *) SUPERLU_MALLOC(p->A.nrow * sizeof(double));
	ReturnErrIf(p->R == NULL);
	p->C = (double *) SUPERLU_MALLOC(p->A.ncol * sizeof(double));
	ReturnErrIf(p->C == NULL);
	p->ferr = (double *) SUPERLU_MALLOC(1 * sizeof(double));
	ReturnErrIf(p->ferr == NULL);
	p->berr = (double *) SUPERLU_MALLOC(1 * sizeof(double));
	ReturnErrIf(p->berr == NULL);
	
	/* Set the default input options. */
	p->options.Fact = DOFACT;
	p->options.Equil = NO;
	p->options.ColPerm = COLAMD;
	p->options.Trans = NOTRANS;
	p->options.IterRefine = NOREFINE;
	p->options.PrintStat = NO;
	p->options.SymmetricMode = NO;
	p->options.DiagPivotThresh = 1.0;
	p->options.PivotGrowth = NO;
	p->options.ConditionNumber = NO;
	
	/* Initialize the statistics variables. */
    StatInit(&p->stat);
	
	return 0;
}
