/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <log.h>
#include "node.h"

/*****************************************************************************/

struct _node {
	int row;
	int col;
	double *data;
};

static double gndNodeData = 0.0;
node_ gndNode = {
	.row = 0,
	.col = 0,
	.data = &gndNodeData
};

int nodeDataPlus(node_ *r, double plus)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->data == NULL);
	*r->data += plus;
	return 0;
}

int nodeDataMinus(node_ *r, double minus)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->data == NULL);
	*r->data -= minus;
	return 0;
}

int nodeDestroy(node_ **r)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	
	Debug("Destroying Node %p", (*r));
	
	free(*r);
	*r = NULL;
	return 0;
}

node_ * nodeNew(node_ *r, int row, int col)
{
	ReturnNULLIf(r != NULL);
	r = calloc(1, sizeof(node_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Node %i:%i %p", row, col, r);
	
	r->row = row;
	r->col = col;
	
	return r;
}

/*****************************************************************************/

struct _nodeDB {
	node_ *node;
	nodeDB_ *next;
};

nodeDB_ * nodeDBNext(nodeDB_ *r)
{
	if(r == NULL) return NULL;
	return r->next;
}

node_ * nodeDBGetNode(nodeDB_ *r)
{
	ReturnNULLIf(r == NULL);
	return r->node;
}

int nodeDBLength(nodeDB_ *r)
{
	int i;
	ReturnErrIf(r == NULL);
	/* TODO: Make this more effecient, add an counter varible */
	/* Don't count the ground node */
	for(i = 0; r->next != NULL; i++, r = r->next);
	return i;
}

int nodeDBInit(nodeDB_ *r, double *data, int *row, int *colStart)
{
	nodeDB_ *ptr;
	int i;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(data == NULL);
	ReturnErrIf(row == NULL);
	ReturnErrIf(colStart == NULL);
	
	for(ptr = r->next, i = 0; ptr != NULL; ptr = ptr->next, i++) {
		ReturnErrIf(ptr->node == NULL);
		ptr->node->data = &data[i];
		row[i] = ptr->node->row - 1;
		colStart[ptr->node->col] = i + 1;
	}
	
	return 0;
}

/* TODO: This could be made faster */
node_ * nodeDBFind(nodeDB_ *r, int row, int col)
{
	nodeDB_ *ptr = NULL, *newElement = NULL;
	
	ReturnNULLIf(r == NULL);
	ReturnNULLIf(r->node == NULL);
	ReturnNULLIf(row < 0);
	ReturnNULLIf(col < 0);
	
	/* node list is sorted by column and then row */
	/* is it the ground node ? */
	if((gndNode.row == row) || (gndNode.col == col)) {
		return &gndNode;
	}
	for(ptr = r; ptr->next != NULL; ptr = ptr->next) {
		ReturnNULLIf(ptr->next->node == NULL);
		if((ptr->next->node->row == row) && (ptr->next->node->col == col))
			return ptr->next->node;
		if((ptr->next->node->col > col) || ((ptr->next->node->col == col) && 
				(ptr->next->node->row > row)))
			break;
	}
	
	/* If can't find it add it */
	newElement = calloc(1, sizeof(nodeDB_));
	ReturnNULLIf(newElement == NULL);
	newElement->node =  nodeNew(newElement->node, row, col);
	ReturnNULLIf(newElement->node == NULL);
	newElement->next = ptr->next;
	ptr->next = newElement;
	
	return newElement->node;
}

int nodeDBAttach(nodeDB_ *r, node_ *node)
{
	nodeDB_ *ptr;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(node == NULL);
	
	for(ptr = r; ptr->next != NULL; ptr = ptr->next);
	
	ptr->next = calloc(1, sizeof(nodeDB_));
	ReturnErrIf(ptr->next == NULL);
	ptr->next->node =  node;
	
	return 0;
}

int nodeDBDestroySafe(nodeDB_ **r)
{
	nodeDB_ *ptr, *ptrNext;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying Node DB %p", (*r));
	
	for(ptr = (*r); ptr != NULL; ptr = ptrNext) {
		ptrNext = ptr->next;
		free(ptr);
	}
	
	return 0;
}

int nodeDBDestroy(nodeDB_ **r)
{
	nodeDB_ *ptr, *ptrNext;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying Node DB %p", (*r));
	
	for(ptr = (*r); ptr != NULL; ptr = ptrNext) {
		ptrNext = ptr->next;
		if(ptr->node != &gndNode) {
			ReturnErrIf(nodeDestroy(&ptr->node));
		}
		free(ptr);
	}
	
	return 0;
}

nodeDB_ * nodeDBNew(nodeDB_ *r)
{
	ReturnNULLIf(r != NULL);
	r = calloc(1, sizeof(nodeDB_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Node DB %p", r);
	/* First node is ground node */
	r->node = &gndNode;
	
	return r;
}

