/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <math.h>
#include <log.h>
#include "row.h"

/*****************************************************************************/

struct _row {
	char *name;
	rowType_ type;
	int index;
	double *rhs;
	double *solution;
};

static double gndRowRHS = 0.0;
static double gndRowSol = 0.0;
row_ gndRow = {
	.name = "0",
	.type = TYPE_NORMAL,
	.index = 0,
	.rhs = &gndRowRHS,
	.solution = &gndRowSol
};

rowType_ rowGetType(row_ *r)
{
	ReturnErrIf(r == NULL);
	return r->type;
}

char * rowGetName(row_ *r)
{
	ReturnNULLIf(r == NULL);
	return r->name;
}

int rowGetIndex(row_ *r)
{
	ReturnErrIf(r == NULL);
	return r->index;
}

double rowGetSolution(row_ *r)
{
	ReturnNaNIf(r == NULL);
	ReturnNaNIf(r->solution == NULL);
	return *r->solution;
}

double ** rowGetSolutionPtr(row_ *r)
{
	ReturnNULLIf(r == NULL);
	return &r->solution;
}


int rowRHSPlus(row_ *r, double plus)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->rhs == NULL);
	*r->rhs += plus;
	return 0;
}

int rowRHSMinus(row_ *r, double minus)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->rhs == NULL);
	*r->rhs -= minus;
	return 0;
}

int rowDestroy(row_ **r)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	
	Debug("Destroying Row %s:%i %p", (*r)->name, (*r)->index, (*r));
	
	free(*r);
	*r = NULL;
	return 0;
}

row_ * rowNew(row_ *r, char *name, rowType_ type, int index)
{
	char *typeName;
	
	ReturnNULLIf(r != NULL);
	ReturnNULLIf(name == NULL);
	ReturnNULLIf(index < 0);
	switch(type) {
		case TYPE_EXTENDED: typeName = "Extended"; break;
		case TYPE_NORMAL: typeName = "Normal"; break;
		default: ReturnNULL("Unknown row type");
	}
	
	r = calloc(1, sizeof(row_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Row %s:%s %p ", name, typeName, r);
	r->name = name;
	r->index = index;
	r->type = type;
	
	return r;
}

/*****************************************************************************/

struct _rowDB {
	row_ *row;
	rowDB_ *next;
};

rowDB_ * rowDBNext(rowDB_ *r)
{
	if(r == NULL) return NULL;
	return r->next;
}

row_ * rowDBGetRow(rowDB_ *r)
{
	ReturnNULLIf(r == NULL);
	return r->row;
}

int rowDBWriteRaw(rowDB_ *r, FILE *outFile)
{
	rowDB_ *ptrRow;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(outFile == NULL);
	
	for(ptrRow = r->next; ptrRow != NULL; ptrRow = ptrRow->next) {
		ReturnErrIf(ptrRow->row == NULL);
		if(ptrRow->row->type == TYPE_NORMAL) {
			fprintf(outFile, "\t%i\tV(%s)\tvoltage\n", ptrRow->row->index, 
					ptrRow->row->name);
		} else if(ptrRow->row->type == TYPE_EXTENDED) {
			if(strchr(ptrRow->row->name, '#') == NULL) {
				fprintf(outFile, "\t%i\t%s#branch\tcurrent\n", 
						ptrRow->row->index, ptrRow->row->name);
			} else {
				fprintf(outFile, "\t%i\t%s\tcurrent\n", 
						ptrRow->row->index, ptrRow->row->name);
			}
		}
	}
	return 0;
}

int rowDBLength(rowDB_ *r)
{
	int i;
	ReturnErrIf(r == NULL);
	/* TODO: Make this more effecient, add an counter varible */
	/* Don't count ground row */
	for(i = 0; r->next != NULL; i++, r = r->next);
	return i;
}

int rowDBInit(rowDB_ *r, double *rhs, double *solution)
{
	rowDB_ *ptr;
	int i;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(rhs == NULL);
	ReturnErrIf(solution == NULL);
	
	for(ptr = r->next, i = 0; ptr != NULL; ptr = ptr->next, i++) {
		ReturnErrIf(ptr->row == NULL);
		ptr->row->rhs = &rhs[i];
		ptr->row->solution = &solution[i];
	}
	
	return 0;
}

/* TODO: This could be made faster */
row_ * rowDBFind(rowDB_ *r, char *name, rowType_ type)
{
	rowDB_ *ptr, *newElement;
	
	ReturnNULLIf(r == NULL);
	ReturnNULLIf(r->row == NULL);
	ReturnNULLIf(name == NULL);
	
	/* is it the ground node ? */
	if(!strcmp(gndRow.name, name)) {
		return &gndRow;
	}
	
	for(ptr = r; ptr->next != NULL; ptr = ptr->next) {
		ReturnNULLIf(ptr->next->row == NULL);
		if((!strcmp(ptr->next->row->name, name)) && 
				(ptr->next->row->type == type)) {
			return ptr->next->row;
		}
	}
	
	ReturnNULLIf(ptr->row == NULL);
	
	/* If can't find it create it and add it */
	newElement = calloc(1, sizeof(rowDB_));
	ReturnNULLIf(newElement == NULL);
	newElement->row =  rowNew(newElement->row, name, type, ptr->row->index + 1);
	ReturnNULLIf(newElement->row == NULL);
	newElement->next = ptr->next;
	ptr->next = newElement;
	
	return newElement->row;
}

int rowDBAttach(rowDB_ *r, row_ *row)
{
	rowDB_ *ptr;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(row == NULL);
	
	for(ptr = r; ptr->next != NULL; ptr = ptr->next) {
		if(ptr->next->row == row) {
			return 1;
		}
	}
	
	/* If can't find it add it */
	ptr->next = calloc(1, sizeof(rowDB_));
	ReturnErrIf(ptr->next == NULL);
	ptr->next->row =  row;
	
	return 0;
}

int rowDBDestroySafe(rowDB_ **r)
{
	rowDB_ *ptr, *ptrNext;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying Row DB %p", (*r));
	
	for(ptr = (*r); ptr != NULL; ptr = ptrNext) {
		ptrNext = ptr->next;
		free(ptr);
	}
	
	return 0;
}

int rowDBDestroy(rowDB_ **r)
{
	rowDB_ *ptr, *ptrNext;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying Row DB %p", (*r));
	
	for(ptr = (*r); ptr != NULL; ptr = ptrNext) {
		ptrNext = ptr->next;
		if(ptr->row != &gndRow) {
			ReturnErrIf(rowDestroy(&ptr->row));
		}
		free(ptr);
	}
	
	return 0;
}

rowDB_ * rowDBNew(rowDB_ *r)
{
	ReturnNULLIf(r != NULL);
	r = calloc(1, sizeof(rowDB_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Row DB %p", r);
	/* First row is ground row */
	r->row = &gndRow;
	
	return r;
}

