/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <log.h>
#include "subckt.h"

/*****************************************************************************/

typedef struct _subcktList subcktList_;
struct _subcktList {
	char *name;
	char **nodes;
	int numNodes;
	char *buffer;
	subcktList_ *next;
};

static subcktList_ * subcktListFindSubckt(subcktList_ *r, char *name)
{
	subcktList_ *ptr;
	
	ReturnNULLIf(r == NULL);
	ReturnNULLIf(r->name == NULL);
	ReturnNULLIf(name == NULL);
	
	for(ptr = r; ptr != NULL; ptr = ptr->next) {
		if(!strcmp(ptr->name, name)) {
			return ptr;
		}
	}
	
	return NULL;
}

static int subcktListFindNodeIndex(subcktList_ *r, char *node)
{
	int i;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->nodes == NULL);
	ReturnErrIf(r->numNodes < 0);
	ReturnErrIf(node == NULL);
	
	for(i = 0; i < r->numNodes; i++) {
		ReturnErrIf(r->nodes[i] == NULL);
		if(!strcmp(r->nodes[i], node)) {
			return i;
		}
	}
	
	return r->numNodes;
}

static int subcktListDestroy(subcktList_ **r)
{
	subcktList_ *ptr, *ptrNext;
	
	ReturnErrIf(r == NULL);
	Debug("Destroying Subckt List %p", (*r));
	
	for(ptr = (*r); ptr != NULL; ptr = ptrNext) {
		ptrNext = ptr->next;
		if(ptr->nodes != NULL) {
			free(ptr->nodes);
		}
		free(ptr);
	}
	
	*r = NULL;
	return 0;
}

static subcktList_ * subcktListNew(subcktList_ *r, char *name, char **nodes, 
		int numNodes, char *buffer)
{
	ReturnNULLIf(r != NULL);
	ReturnNULLIf(name == NULL);
	ReturnNULLIf(nodes == NULL);
	ReturnNULLIf(numNodes < 1);
	ReturnNULLIf(buffer == NULL);
	r = calloc(1, sizeof(subcktList_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Subckt List %p", r);
	
	r->name = name;
	r->nodes = nodes;
	r->numNodes = numNodes;
	r->buffer = buffer;
	
	return r;
}

/*****************************************************************************/

typedef struct _stringList stringList_;
struct _stringList {
	char *string;
	stringList_ *next;
};

static int stringListDestroy(stringList_ **r)
{
	stringList_ *ptr, *ptrNext;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying Name DB %p", (*r));
	
	for(ptr = (*r); ptr != NULL; ptr = ptrNext) {
		ptrNext = ptr->next;
		if(ptr->string != NULL) {
			free(ptr->string);
		}
		free(ptr);
	}
	
	return 0;
}

static stringList_ * stringListNew(stringList_ *r)
{
	ReturnNULLIf(r != NULL);
	r = calloc(1, sizeof(stringList_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating String List %p", r);
	
	return r;
}

static int stringListAddString(stringList_ *r, char *string)
{
	stringList_ *stringList = NULL;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(string == NULL);
	
	stringList = stringListNew(stringList);
	ReturnErrIf(stringList == NULL);
	
	stringList->string = string;
	
	stringList->next = r->next;
	r->next = stringList;
	
	return 0;
}

/*****************************************************************************/

typedef struct _xElement xElement_;
struct _xElement {
	char *name;
	char **nodes;
	int numNodes;
	char *subcktName;
	char *buffer;
	subcktList_ *subckt;
	xElement_ *next;
};

static int xElementDestroy(xElement_ **r)
{
	xElement_ *ptr, *ptrNext;
	
	ReturnErrIf(r == NULL);
	Debug("Destroying X DB %p", (*r));
	
	for(ptr = (*r); ptr != NULL; ptr = ptrNext) {
		ptrNext = ptr->next;
		if(ptr->nodes != NULL) {
			free(ptr->nodes);
		}
		if(ptr->buffer != NULL) {
			free(ptr->buffer);
		}
		free(ptr);
	}
	*r = NULL;
	
	return 0;
}

static xElement_ * xElementNew(xElement_ *r, char *name, char **nodes, 
	int numNodes, char *subcktName)
{
	ReturnNULLIf(r != NULL);
	ReturnNULLIf(name == NULL);
	ReturnNULLIf(nodes == NULL);
	ReturnNULLIf(numNodes < 1);
	ReturnNULLIf(subcktName == NULL);
	r = calloc(1, sizeof(xElement_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating X DB %p %s", r, name);
	r->name = name;
	r->nodes = nodes;
	r->numNodes = numNodes;
	r->subcktName = subcktName;
	
	return r;
}

/*****************************************************************************/

struct _subcktDB {
	subcktList_ *subckts;
	xElement_ *xsHead;
	xElement_ *xsTail;
	xElement_ *xsCursor;
	stringList_ *strings;
};

char * subcktDBTransformRefDes(subcktDB_ *r, char *refdes)
{
	char *newRefdes;
	
	ReturnNULLIf(r == NULL);
	ReturnNULLIf(refdes == NULL);
	ReturnNULLIf(r->xsCursor == NULL);
	ReturnNULLIf(r->xsCursor->name == NULL);
	
	newRefdes = malloc(strlen(refdes)+strlen(r->xsCursor->name) + 2);
	ReturnNULLIf(newRefdes == NULL);
	
	sprintf(newRefdes, "%s:%s", r->xsCursor->name, refdes);
	ReturnNULLIf(stringListAddString(r->strings, newRefdes));
	
	Debug("Mapping %s to %s", refdes, newRefdes);
	
	return newRefdes;
}

char * subcktDBTransformNode(subcktDB_ *r, char *node)
{
	char *newNode;
	int i;
	
	ReturnNULLIf(r == NULL);
	ReturnNULLIf(node == NULL);
	ReturnNULLIf(r->xsCursor == NULL);
	ReturnNULLIf(r->xsCursor->nodes == NULL);
	ReturnNULLIf(r->xsCursor->numNodes < 0);
	
	i = subcktListFindNodeIndex(r->xsCursor->subckt, node);
	ReturnNULLIf(i < 0);
	if(i < r->xsCursor->numNodes) {
		ReturnNULLIf(r->xsCursor->nodes[i] == NULL);
		Debug("Mapping %s to %s", node, r->xsCursor->nodes[i]);
		return r->xsCursor->nodes[i];
	}
	
	newNode = malloc(strlen(node)+strlen(r->xsCursor->name) + 2);
	ReturnNULLIf(newNode == NULL);
	
	sprintf(newNode, "%s:%s", r->xsCursor->name, node);
	ReturnNULLIf(stringListAddString(r->strings, newNode));
	
	Debug("Mapping %s to %s", node, newNode);
	
	return newNode;
}

subcktXListStatus subcktDBNextX(subcktDB_ *r)
{
	xElement_ *x;
	
	ReturnErrIf(r == NULL);
	if(r->xsHead == NULL)
		return X_DONE;
	
	if(r->xsCursor == NULL) {
		x = r->xsHead;
	} else if (r->xsCursor == r->xsTail){
		return X_DONE;
	} else {
		x = r->xsCursor->next;
	}
	
	if(x->subckt == NULL) {
		x->subckt = subcktListFindSubckt(r->subckts, x->subcktName);
		ReturnErrIf(x->subckt == NULL);
		ReturnErrIf(x->subckt->numNodes != x->numNodes);
		x->buffer = malloc(strlen(x->subckt->buffer)+1);
		ReturnErrIf(x->buffer == NULL);
		strcpy(x->buffer, x->subckt->buffer);
	}
	
	r->xsCursor = x;
	return 0;
}

char * subcktDBGetActiveBuffer(subcktDB_ *r)
{
	ReturnNULLIf(r == NULL);
	ReturnNULLIf(r->xsCursor == NULL);
	ReturnNULLIf(r->xsCursor->subckt == NULL);
	ReturnNULLIf(r->xsCursor->subckt->buffer == NULL);
	return r->xsCursor->buffer;
}

int subcktDBAddXElement(subcktDB_ *r, char *name, char **nodes, int numNodes, 
		char *subcktName)
{
	xElement_ *xElement = NULL;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(name == NULL);
	ReturnErrIf(nodes == NULL);
	ReturnErrIf(numNodes < 1);
	ReturnErrIf(subcktName == NULL);
	
	xElement = xElementNew(xElement, name, nodes, numNodes, subcktName);
	ReturnErrIf(xElement == NULL);
	
	if(r->xsHead == NULL) {
		r->xsHead = xElement;
		r->xsTail = xElement;
	} else {
		r->xsTail->next = xElement;
		r->xsTail = xElement;
	}
	
	return 0;
}

int subcktDBAddSubckt(subcktDB_ *r, char *name, char **nodes, int numNodes, 
		char *buffer)
{
	subcktList_ *subcktList = NULL;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(name == NULL);
	ReturnErrIf(nodes == NULL);
	ReturnErrIf(numNodes < 1);
	ReturnErrIf(buffer == NULL);
	
	Debug("Adding subckt %s", name);
	
	subcktList = subcktListNew(subcktList, name, nodes, numNodes, buffer);
	ReturnErrIf(subcktList == NULL);
	
	subcktList->next = r->subckts;
	r->subckts = subcktList;
	
	return 0;
}

int subcktDBDestroy(subcktDB_ **r)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying Subckt DB %p", (*r));
	
	ReturnErrIf(xElementDestroy(&(*r)->xsHead));
	ReturnErrIf(subcktListDestroy(&(*r)->subckts));
	ReturnErrIf(stringListDestroy(&(*r)->strings));
	
	return 0;
}

subcktDB_ * subcktDBNew(subcktDB_ *r)
{
	ReturnNULLIf(r != NULL);
	r = calloc(1, sizeof(subcktDB_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Subckt DB %p", r);
	
	r->strings = stringListNew(r->strings);
	ReturnNULLIf(r->strings == NULL);
	
	return r;
}

