/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <math.h>
#include <log.h>
#include "device_internal.h"

static char *type = "Capacitor";

#define NUM_PINS	2
#define K			0
#define J			1

#define N	3
#define M	2

struct _devicePrivate {
	double C;
	double ic;
	double G[M];
	double Ieq[M];
	double T[N];
	double h[N];
	double V[N];
	double I[N];
	int n;
	int m;
	row_ *rowK;
	row_ *rowJ;
	node_ *nodeKK;
	node_ *nodeJK;
	node_ *nodeKJ;
	node_ *nodeJJ;
};

static double nextStep(device_ *r, state_ *state, deviceNIType niType)
{
	devicePrivate_ *p;
	double v, i, exp, ex, e, dd, Ch[4], h = 0;
	
	ReturnNaNIf(r == NULL);
	ReturnNaNIf(state == NULL);
	p = r->private;
	ReturnNaNIf(p == NULL);
	
	/* Get voltage and current */
	v = rowGetSolution(p->rowK) - rowGetSolution(p->rowJ);
	ReturnNaNIf(isnan(v));
	i =  p->G[(p->m-1)%M] * v - p->Ieq[(p->m-1)%M];
	
	/* Calculate Charge */
	Ch[0] = p->C * v;
	Ch[1] = p->C * p->V[(p->n)%N];
	Ch[2] = p->C * p->V[(p->n-1)%N];
	Ch[3] = p->C * p->V[(p->n-2)%N];
	
	/* Calculate current error */
	exp = stateGetAbsError(state, Max2(fabs(p->I[p->n%N]), fabs(i)));
	ReturnNaNIf(isnan(exp));
	
	/* Calculate charge error (converted to current by dividing by h)*/
	ex = stateGetChgError(state, Max2(fabs(Ch[0]), fabs(Ch[1]))) / p->h[p->n%N];
	ReturnNaNIf(isnan(ex));
	
	/* Take maximum current */
	e = Max2(ex, exp);
	
	switch(niType) {
	case NI_ERR:
		ReturnNaNIf("Unknown NI Type");
	case NI_BE:
		dd = DD2(Ch[0], Ch[1], Ch[2], p->h[p->n%N], p->h[(p->n-1)%N]);
		h = stateGetTrAbsError(state, dd/2, e);
		ReturnNaNIf(isnan(h));
		break;
	case NI_TRAP:
		dd = DD3(Ch[0], Ch[1], Ch[2], Ch[3], p->h[p->n%N], p->h[(p->n-1)%N],
			p->h[(p->n-2)%N]);
		h = sqrtf(stateGetTrAbsError(state, dd/12, e));
		ReturnNaNIf(isnan(h));
		break;
	}
	
	return h;
}

static int step(device_ *r, state_ *state, double time, deviceNIType niType)
{
	devicePrivate_ *p;
	double gDelta, iDelta;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(state == NULL);
	ReturnErrIf(time < 0.0);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	/* Modified Nodal Analysis Stamp
	 *				          +    Gn    -
	 *	  |_Vk__Vj_|_rhs_|	  +--/\/\/\--+
	 *	k | Gn -Gn |-Ieq |	k_|    __    |_j
	 *	j | -Gn Gn | Ieq |	  |__ /  \___|
	 *	   					      \__/
	 *	        				   Ieq
	 */
	
	if(time > p->T[(p->n+1)%N]) {
		p->n++;
	}
	
	/* Set New Time */
	p->h[p->n%N] = time - p->T[p->n%N];
	p->T[(p->n+1)%N] = time;
	
	/* Record current and voltage */
	p->V[p->n%N] = rowGetSolution(p->rowK) - rowGetSolution(p->rowJ);
	ReturnErrIf(isnan(p->V[MOD(p->n,N)]));
	p->I[p->n%N] = p->G[(p->m-1)%M] * p->V[p->n%N] - p->Ieq[(p->m-1)%M];
	
	/* Calculate Next Voltage USing Numerical Integration */
	switch(niType) {
	case NI_ERR:
		ReturnErr("Unknown NI Type");
	case NI_BE:
		p->G[p->m%M] = p->C / p->h[p->n%N];
		p->Ieq[p->m%M] = p->G[p->m%M] * p->V[p->n%N];
		break;
	case NI_TRAP:
		p->G[p->m%M] = 2 * p->C / p->h[p->n%N];
		p->Ieq[p->m%M] = p->G[p->m%M] * p->V[p->n%N] + p->I[p->n%N];
		break;
	}
	
	/* Set-up Matrices */
	gDelta = p->G[p->m%M] - p->G[(p->m-1)%M];
	ReturnErrIf(nodeDataPlus(p->nodeKK, gDelta));
	ReturnErrIf(nodeDataMinus(p->nodeKJ, gDelta));
	ReturnErrIf(nodeDataMinus(p->nodeJK, gDelta));
	ReturnErrIf(nodeDataPlus(p->nodeJJ, gDelta));
	iDelta = p->Ieq[p->m%M] - p->Ieq[(p->m-1)%M];
	ReturnErrIf(rowRHSPlus(p->rowK, iDelta));
	ReturnErrIf(rowRHSMinus(p->rowJ, iDelta));
	p->m++;
	
	return 0;
}

static int load(device_ *r, state_ *state)
{
	devicePrivate_ *p;
	ReturnErrIf(r == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	Debug("%s Load %p", type, r);
	
	/* Modified Nodal Analysis Stamp (Open)
	 *	                  	+      -
	 *	  |_Vk_Vj_|_rhs_|	--o  o--
	 *	k | -- -- | --  |	k      j
	 *	j | -- -- | --  |
	 */
	
	return 0;
}

int deviceConfigCapacitor(device_ *r, double C, double ic)
{
	devicePrivate_ *p;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->private != NULL);
	ReturnErrIf(r->numPins != NUM_PINS);
	
	Debug("Configuring %s %p", type, r);
	
	r->type = type;
	r->load = load;
	r->step = step;
	r->nextStep = nextStep;
	
	r->private =  calloc(1, sizeof(devicePrivate_));
	ReturnErrIf(r->private == NULL);
	p = r->private;
	
	p->C = C;
	p->ic = ic;
	
	p->rowK = r->pin[K];
	ReturnErrIf(p->rowK == NULL);
	p->rowJ = r->pin[J];
	ReturnErrIf(p->rowJ == NULL);
	
	p->nodeKK = matrixFindNode(r->matrix, p->rowK, p->rowK);
	ReturnErrIf(p->nodeKK == NULL);
	p->nodeJK = matrixFindNode(r->matrix, p->rowJ, p->rowK);
	ReturnErrIf(p->nodeJK == NULL);
	p->nodeKJ = matrixFindNode(r->matrix, p->rowK, p->rowJ);
	ReturnErrIf(p->nodeKJ == NULL);
	p->nodeJJ = matrixFindNode(r->matrix, p->rowJ, p->rowJ);
	ReturnErrIf(p->nodeJJ == NULL);
	
	return 0;
}
