/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <math.h>
#include <log.h>
#include "device_internal.h"

static char *type = "Inductor";

#define NUM_PINS	2
#define K			0
#define J			1

#define N	3
#define M	2

struct _devicePrivate {
	double L;
	double ic;
	double R[M];
	double Veq[M];
	double T[N];
	double h[N];
	double V[N];
	double I[N];
	int n;
	int m;
	row_ *rowK;
	row_ *rowJ;
	row_ *rowR;
	node_ *nodeRK;
	node_ *nodeRJ;
	node_ *nodeKR;
	node_ *nodeJR;
	node_ *nodeRR;
};

static double nextStep(device_ *r, state_ *state, deviceNIType niType)
{
	devicePrivate_ *p;
	double v, i, exp, ex, e, dd, Fl[4], h = 0;
	
	ReturnNaNIf(r == NULL);
	ReturnNaNIf(state == NULL);
	p = r->private;
	ReturnNaNIf(p == NULL);
	
	/* Get voltage and current */
	v = rowGetSolution(p->rowK) - rowGetSolution(p->rowJ);
	ReturnNaNIf(isnan(v));
	i = rowGetSolution(p->rowR);
	ReturnNaNIf(isnan(i));
	
	/* Calculate Flux */
	Fl[0] = p->L * i;
	Fl[1] = p->L * p->I[(p->n)%N];
	Fl[2] = p->L * p->I[(p->n-1)%N];
	Fl[3] = p->L * p->I[(p->n-2)%N];
	
	/* Calculate voltage error */
	exp = stateGetVnError(state, Max2(fabs(p->V[p->n%N]), fabs(v)));
	ReturnNaNIf(isnan(exp));
	/* Calculate flux error (converted to voltage by dividing by h)*/
	ex = stateGetChgError(state, Max2(fabs(Fl[0]), fabs(Fl[1]))) / p->h[p->n%N];
	ReturnNaNIf(isnan(ex));
	
	/* Take maximum volatge */
	e = Max2(ex, exp);
	
	switch(niType) {
	case NI_ERR:
		ReturnNaNIf("Unknown NI Type");
	case NI_BE:
		dd = DD2(Fl[0], Fl[1], Fl[2], p->h[p->n%N], p->h[(p->n-1)%N]);
		h = stateGetTrVnError(state, dd/2, e);
		ReturnNaNIf(isnan(h));
		break;
	case NI_TRAP:
		dd = DD3(Fl[0], Fl[1], Fl[2], Fl[3], p->h[p->n%N], p->h[(p->n-1)%N],
			p->h[(p->n-2)%N]);
		h = sqrt(stateGetTrVnError(state, dd/12, e));
		ReturnNaNIf(isnan(h));
		break;
	}
	
	return h;
}

static int step(device_ *r, state_ *state, double time, deviceNIType niType)
{
	devicePrivate_ *p;
	double rDelta, vDelta;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(state == NULL);
	ReturnErrIf(time < 0.0);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	/* Modified Nodal Analysis Stamp
	 *				         
	 *	  |_Vk_Vj_Ir_|_rhs_|	 
	 *	k | -- --  1 | --  |	 +    /\     Rn    -
	 *	j | -- -- -1 | --  |	 k__ /  \__/\/\/\__j 
	 *	r | 1  -1 -Rn| Veq |	     \  /
	 *	        				      \/ Veq
	 */
	
	if(time > p->T[(p->n+1)%N]) {
		p->n++;
	}
	
	/* Set New Time */
	p->h[p->n%N] = time - p->T[p->n%N];
	p->T[(p->n+1)%N] = time;
	
	/* Record current and voltage */
	p->V[p->n%N] = rowGetSolution(p->rowK) - rowGetSolution(p->rowJ);
	ReturnErrIf(isnan(p->V[MOD(p->n,N)]));
	p->I[p->n%N] = rowGetSolution(p->rowR);
	ReturnErrIf(isnan(p->I[MOD(p->n,N)]));
	
	/* Calculate Next Voltage base on Numerical Integration */
	switch(niType) {
	case NI_ERR:
		ReturnErr("Unknown NI Type");
	case NI_BE:
		p->R[p->m%M] = p->L / p->h[p->n%N];
		p->Veq[p->m%M] = p->R[p->m%M] * p->I[p->n%N];
		break;
	case NI_TRAP:
		p->R[p->m%M] = 2 * p->L / p->h[p->n%N];
		p->Veq[p->m%M] = p->R[p->m%M] * p->I[p->n%N] + p->V[p->n%N];
		break;
	}
	
	/* Set-up Matrices */
	rDelta = (p->R[p->m%M] - p->R[(p->m-1)%M]);
	ReturnErrIf(nodeDataMinus(p->nodeRR, rDelta));
	vDelta = p->Veq[p->m%M] - p->Veq[(p->m-1)%M];
	ReturnErrIf(rowRHSMinus(p->rowR, vDelta));
	p->m++;
	
	return 0;
}

static int load(device_ *r, state_ *state)
{
	devicePrivate_ *p;
	ReturnErrIf(r == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	Debug("%s Load %p", type, r);
	
	/* Modified Nodal Analysis Stamp (Short)
	 *	                     	+      -
	 *	  |_Vk_Vj_Ir_|_rhs_|	________
	 *	k | -- --  1 | --  |	k      j
	 *	j | -- -- -1 | --  |	     
	 *	r |  1 -1 -- | --  |	------->
	 *	                    	   Ir
	 */
	
	ReturnErrIf(nodeDataPlus(p->nodeRK, 1.0));
	ReturnErrIf(nodeDataMinus(p->nodeRJ, 1.0));
	ReturnErrIf(nodeDataPlus(p->nodeKR, 1.0));
	ReturnErrIf(nodeDataMinus(p->nodeJR, 1.0));	
	
	return 0;
}

int deviceConfigInductor(device_ *r, double L, double ic)
{
	devicePrivate_ *p;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->private != NULL);
	ReturnErrIf(r->numPins != NUM_PINS);
	
	Debug("Configuring %s %p", type, r);
	
	r->type = type;
	r->load = load;
	r->step = step;
	r->nextStep = nextStep;
	
	r->private =  calloc(1, sizeof(devicePrivate_));
	ReturnErrIf(r->private == NULL);
	p = r->private;
	
	p->L = L;
	p->ic = ic;
	
	p->rowR = matrixFindRow(r->matrix, r->refdes, TYPE_EXTENDED);
	ReturnErrIf(p->rowR == NULL);
	p->rowK = r->pin[K];
	ReturnErrIf(p->rowK == NULL);
	p->rowJ = r->pin[J];
	ReturnErrIf(p->rowJ == NULL);
	
	p->nodeRK = matrixFindNode(r->matrix, p->rowR, p->rowK);
	ReturnErrIf(p->nodeRK == NULL);
	p->nodeRJ = matrixFindNode(r->matrix, p->rowR, p->rowJ);
	ReturnErrIf(p->nodeRJ == NULL);
	p->nodeKR = matrixFindNode(r->matrix, p->rowK, p->rowR);
	ReturnErrIf(p->nodeKR == NULL);
	p->nodeJR = matrixFindNode(r->matrix, p->rowJ, p->rowR);
	ReturnErrIf(p->nodeJR == NULL);
	p->nodeRR = matrixFindNode(r->matrix, p->rowR, p->rowR);
	ReturnErrIf(p->nodeJR == NULL);
	
	return 0;
}
