/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <math.h>
#include <calc.h>
#include <log.h>
#include "device_internal.h"

static char *type = "Non-linear Dependent Source";

#define NUM_PINS	2
#define K			0
#define J			1

struct _devicePrivate {
	double gmin;
	calc_ *calc;
	rowDB_ *rowDB;
	nodeDB_ *nodeDB;
	int numVariables;
	node_ *nodeRK;
	node_ *nodeRJ;
	node_ *nodeKR;
	node_ *nodeJR;
	row_ *rowR;
	row_ *rowK;
	row_ *rowJ;
	double Vp;	/* previous value */
	double Vc; /* calculated value */
	double Vn; /* current value */
	double *diffs; /* differentials */
	int status; /* convergence status */
};

static int calculateVoltages(devicePrivate_ *p)
{
	p->Vn = rowGetSolution(p->rowK) - rowGetSolution(p->rowJ);
	ReturnErrIf(isnan(p->Vn));
	ReturnErrIf(calcSolve(p->calc, &p->Vc));
	return 0;
}

static int checkConvergence(devicePrivate_ *p, state_ *state)
{
	double e;
	e = stateGetVnError(state, Max2(fabs(p->Vp), fabs(p->Vn)));
	ReturnErrIf(isnan(e));
	if(fabs(p->Vp - p->Vn) <= e) {
		e = stateGetVnError(state, Max2(fabs(p->Vc), fabs(p->Vn)));
		ReturnErrIf(isnan(e));
		if((p->status > 0) && (fabs(p->Vn - p->Vc) <= e)) {
			p->Vp = p->Vn;
			p->status = 2;
		} else {
			p->status = 1;
		}
	} else {
		p->status = 0;
	}
	return 0;
}

static int loadConductances(devicePrivate_ *p)
{
	nodeDB_ *nodePtr;
	rowDB_ *rowPtr;
	row_ *row;
	rowType_ rowType;
	double dc = 0, diff;
	char iv = 0;
	int i;
	
	nodePtr = nodeDBNext(p->nodeDB);
	rowPtr = rowDBNext(p->rowDB);
	i = 1;
	dc = p->Vc;
	while((nodePtr != NULL) && (rowPtr != NULL)) {
		row = rowDBGetRow(rowPtr);
		ReturnErrIf(row == NULL);
		rowType = rowGetType(row);
		switch(rowType) {
			case TYPE_EXTENDED: iv = 'i'; break;
			case TYPE_NORMAL: iv = 'v'; 	break;
			case TYPE_NONE: ReturnErr("Failed to get row type");
		}
		ReturnErrIf(calcDiff(p->calc, rowGetName(row), iv, &diff));
		ReturnErrIf(nodeDataMinus(nodeDBGetNode(nodePtr), 
				(diff - p->diffs[i])));
		dc -= diff*rowGetSolution(row);
		ReturnErrIf(isnan(dc));
		p->diffs[i] = diff;
		nodePtr = nodeDBNext(nodePtr);
		rowPtr = rowDBNext(rowPtr);
		i++;
	}
	
	ReturnErrIf(rowRHSPlus(p->rowR, dc - p->diffs[0]));
	p->diffs[0] = dc;
	p->Vp = p->Vn;
	return 0;
}

static int linearize(device_ *r, state_ *state)
{
	devicePrivate_ *p;
	ReturnErrIf(r == NULL);
	ReturnErrIf(state == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	Debug("%s Linearize %s", type, r->refdes);
	
	ReturnErrIf(calculateVoltages(p));
	
	ReturnErrIf(checkConvergence(p, state));
	
	if(p->status == 2) {
		return LIN_LINEAR;
	} else {
		ReturnErrIf(loadConductances(p));
		return LIN_NOT_LINEAR;
	}
}

static int load(device_ *r, state_ *state)
{
	devicePrivate_ *p;
	ReturnErrIf(r == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	Debug("%s Load %p", type, r);
	
	p->gmin = stateGetGmin(state);
	ReturnErrIf(isnan(p->gmin));
	
	/* Modified Nodal Analysis Stamp
	 *	                     	        +  /\  -
	 *	  |_Vk_Vj_Ir_Vpx_Isy_|_rhs_|	__/Vr\__
	 *	k | -- --  1 --  --  | --  |	k \  / j
	 *	j | -- -- -1 --  --  | --  |	   \/ Vr = f(Vp0, Vp1, ..., Is0, Is1)
	 *	r |  1 -1 -- -gx -gy | Vr  |	------->
	 *	                    	   		   Ir
	 *	NOTEs:	- gx is the derivative of the function wrt the volatge Vpx 
	 *			- all voltages are relative to ground
	 *			- Gy is the derivative of the function wrt the current Isy.
	 *			- Vr is the constant value, i.e.: 
	 *				Vr = f(Vx0,Vx1,...,Iy0,Iy1,...)-gx*Vx0-gx*Vx1-gx*Iy0...
	 */
	
	ReturnErrIf(nodeDataPlus(p->nodeRK, 1.0));
	ReturnErrIf(nodeDataMinus(p->nodeRJ, 1.0));
	ReturnErrIf(nodeDataPlus(p->nodeKR, 1.0));
	ReturnErrIf(nodeDataMinus(p->nodeJR, 1.0));
	
	ReturnErrIf(calculateVoltages(p));
	ReturnErrIf(loadConductances(p));
	
	return 0;
}

static int unconfig(device_ *r)
{
	devicePrivate_ *p;
	
	ReturnErrIf(r == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	ReturnErrIf(calcDestroy(&p->calc));
	ReturnErrIf(rowDBDestroySafe(&p->rowDB));
	ReturnErrIf(nodeDBDestroySafe(&p->nodeDB));
	if(p->diffs != NULL) {
		free(p->diffs);
	}
	
	return 0;
}

static double ** getVariable(char *name, char type, void *private)
{
	device_ *r;
	devicePrivate_ *p;
	row_ *row;
	node_ *node;
	int exists;
	double **variable;
	
	r = (device_*)private;
	ReturnNULLIf(r == NULL);
	p = r->private;
	ReturnNULLIf(p == NULL);
	
	ReturnNULLIf(((type != 'i') && (type != 'v')), "Unknown type %c", type);
	
	row = matrixFindRow(r->matrix, name,
			(type == 'i') ? TYPE_EXTENDED : TYPE_NORMAL);
	ReturnNULLIf(row == NULL);
	exists = rowDBAttach(p->rowDB, row);
	ReturnNULLIf(exists == -1);
	if(!exists) {
		node = matrixFindNode(r->matrix, p->rowR, row);
		ReturnNULLIf(node == NULL);
		ReturnNULLIf(nodeDBAttach(p->nodeDB, node));
		p->numVariables++;
	}
	
	variable = rowGetSolutionPtr(row);
	ReturnNULLIf(variable == NULL);
	
	return variable;
}

int deviceConfigNonlinearV(device_ *r, char *equation)
{
	devicePrivate_ *p;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->private != NULL);
	ReturnErrIf(r->numPins != NUM_PINS);
	ReturnErrIf(equation == NULL);
	
	ReturnErrIf((r->pin[K] == &gndRow) && (r->pin[J] == &gndRow));
	
	Debug("Configuring %s %p", type, r);
	
	r->type = type;
	r->load = load;
	r->unconfig = unconfig;
	r->linearize = linearize;
	
	r->private =  calloc(1, sizeof(devicePrivate_));
	ReturnErrIf(r->private == NULL);
	p = r->private;
	
	p->rowR = matrixFindRow(r->matrix, r->refdes, TYPE_EXTENDED);
	ReturnErrIf(p->rowR == NULL);
	p->rowK = r->pin[K];
	ReturnErrIf(p->rowK == NULL);
	p->rowJ = r->pin[J];
	ReturnErrIf(p->rowJ == NULL);
	
	p->nodeRK = matrixFindNode(r->matrix, p->rowR, p->rowK);
	ReturnErrIf(p->nodeRK == NULL);
	p->nodeRJ = matrixFindNode(r->matrix, p->rowR, p->rowJ);
	ReturnErrIf(p->nodeRJ == NULL);
	p->nodeKR = matrixFindNode(r->matrix, p->rowK, p->rowR);
	ReturnErrIf(p->nodeKR == NULL);
	p->nodeJR = matrixFindNode(r->matrix, p->rowJ, p->rowR);
	ReturnErrIf(p->nodeJR == NULL);
	
	p->rowDB = rowDBNew(p->rowDB);
	ReturnErrIf(p->rowDB == NULL);
	p->nodeDB = nodeDBNew(p->nodeDB);
	ReturnErrIf(p->nodeDB == NULL);
	p->numVariables = 1;
	p->calc = calcNew(p->calc, equation, getVariable, r, &p->gmin);
	ReturnErrIf(p->calc == NULL, "Bad B equation: \n%s", equation);
	
	p->diffs = calloc(p->numVariables, sizeof(double));
	ReturnErrIf(p->diffs == NULL);
	
	return 0;
}

