/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <string.h>
#include <math.h>
#include <log.h>
#include "device_internal.h"
#include "result.h"

#define NUM_PINS	4
#define K			0
#define J			1
#define L			2
#define M			3

static char *type = "Lossless Transmission-Line";

struct _devicePrivate {
	double Z0;
	double Td;
	double Vr;
	double Vs;
	resultDB_ *results;
	row_ *rowK;
	row_ *rowJ;
	row_ *rowL;
	row_ *rowM;
	row_ *rowR;
	char *rowRName;
	row_ *rowS;
	char *rowSName;
	node_ *nodeRK;
	node_ *nodeRJ;
	node_ *nodeKR;
	node_ *nodeJR;
	node_ *nodeRL;
	node_ *nodeRM;
	node_ *nodeSL;
	node_ *nodeSM;
	node_ *nodeLS;
	node_ *nodeMS;
	node_ *nodeSK;
	node_ *nodeSJ;
	node_ *nodeRR;
	node_ *nodeSS;
	node_ *nodeKS;
	node_ *nodeJS;
	node_ *nodeLR;
	node_ *nodeMR;
};

static double nextBreak(device_ *r, state_ *state, double time)
{
	devicePrivate_ *p;
	resultDB_ *results;
	double tBrk;
	
	ReturnNaNIf(r == NULL);
	ReturnNaNIf(time < 0.0);
	p = r->private;
	ReturnNaNIf(p == NULL);
	
	results = resultDBNextBreak(p->results, time, &tBrk);
	ReturnNaNIf(results == NULL);
	return tBrk;
	
	return 0;
}

static int step(device_ *r, state_ *state, double time, deviceNIType niType)
{
	devicePrivate_ *p;
	double Vl, Vm, Vj, Vk, Ir, Is, Vs, Vr;
	double tp;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(state == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	/* Modified Nodal Analysis Stamp
	 *	                     		 	     /\     Zo    -
	 *	  |_Vk_Vj_Ir_Vl_Vm_Is_|_rhs_|	k__ /  \__/\/\/\__j
	 *	k | -- --  1 -- -- -- | --  |	    \  /
	 *	j | -- -- -1 -- -- -- | --  |	     \/ Vr = Vlm(t-Td) + Zo*Ilm(t-Td)
	 *	r | 1  -1 -Zo-- -- -- | Vr  |	+    /\     Zo    -
	 *	l | -- -- -- -- --  1 | --  |  l__ /  \__/\/\/\__m
	 *	m | -- -- -- -- -- -1 | --  |      \  /
	 *	s | -- -- --  1 -1 -Zo| Vs  |       \/ Vs = Vkj(t-Td) + Zo*Ikj(t-Td)
	 */
	
	tp = time - p->Td;
	
	if(resultDBPrevTime(p->results) == 0.0) {
		/* Remove explict km -> lm link */  
		ReturnErrIf(nodeDataMinus(p->nodeRL, 1.0));
		ReturnErrIf(nodeDataMinus(p->nodeRM, 1.0));
		ReturnErrIf(nodeDataMinus(p->nodeSK, 1.0));
		ReturnErrIf(nodeDataMinus(p->nodeSJ, 1.0));
		ReturnErrIf(nodeDataMinus(p->nodeKS, 1.0));
		ReturnErrIf(nodeDataMinus(p->nodeJS, 1.0));
		ReturnErrIf(nodeDataMinus(p->nodeLR, 1.0));
		ReturnErrIf(nodeDataMinus(p->nodeMR, 1.0));
	} else if(tp <= 0) {
		/* Do nothing for now, add initial condition support someday */
	} else {
		ReturnErrIf(isnan(tp));
		p->results = resultDBLinterp(p->results, tp, rowGetIndex(p->rowR), &Ir);
		ReturnErrIf(p->results == NULL);
		p->results = resultDBLinterp(p->results, tp, rowGetIndex(p->rowS), &Is);
		ReturnErrIf(p->results == NULL);
		p->results = resultDBLinterp(p->results, tp, rowGetIndex(p->rowK), &Vk);
		ReturnErrIf(p->results == NULL);
		p->results = resultDBLinterp(p->results, tp, rowGetIndex(p->rowJ), &Vj);
		ReturnErrIf(p->results == NULL);
		p->results = resultDBLinterp(p->results, tp, rowGetIndex(p->rowL), &Vl);
		ReturnErrIf(p->results == NULL);
		p->results = resultDBLinterp(p->results, tp, rowGetIndex(p->rowM), &Vm);
		ReturnErrIf(p->results == NULL);
		
		Vr = (Vl - Vm) + p->Z0*Is;
		Vs = (Vk - Vj) + p->Z0*Ir;
		ReturnErrIf(rowRHSPlus(p->rowR, Vr - p->Vr));
		p->Vr = Vr;
		ReturnErrIf(rowRHSPlus(p->rowS, Vs - p->Vs));
		p->Vs = Vs;
	}

	return 0;
}

static int load(device_ *r, state_ *state)
{
	devicePrivate_ *p;
	ReturnErrIf(r == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	Debug("%s Load %p", type, r);
	
	/* Modified Nodal Analysis Stamp
	 *	                     		 	     /\     0     -
	 *	  |_Vk_Vj_Ir__Vl_Vm_Is_|_rhs_|	k__ /  \__/\/\/\__j
	 *	k | -- --  1  -- -- -1 | --  |	    \  /
	 *	j | -- -- -1  -- --  1 | --  |	     \/ Vm = Vlm
	 *	r | 1  -1 -Zo -1  1 -- | --  |	+     /\    0      -
	 *	l | -- -- -1  -- --  1 | --  |  l__ /  \__/\/\/\__m
	 *	m | -- --  1  -- -- -1 | --  |      \  /
	 *	s |  1 -1  -- 1 -1 -Zo | --  |       \/ Vs = Vkj
	 */
	
	ReturnErrIf(nodeDataPlus(p->nodeRK, 1.0));
	ReturnErrIf(nodeDataMinus(p->nodeRJ, 1.0));
	ReturnErrIf(nodeDataPlus(p->nodeKR, 1.0));
	ReturnErrIf(nodeDataMinus(p->nodeJR, 1.0));
	
	ReturnErrIf(nodeDataPlus(p->nodeSL, 1.0));
	ReturnErrIf(nodeDataMinus(p->nodeSM, 1.0));
	ReturnErrIf(nodeDataPlus(p->nodeLS, 1.0));
	ReturnErrIf(nodeDataMinus(p->nodeMS, 1.0));
	
	ReturnErrIf(nodeDataMinus(p->nodeRR, p->Z0));
	ReturnErrIf(nodeDataMinus(p->nodeSS, p->Z0));
	
	ReturnErrIf(nodeDataPlus(p->nodeRL, 1.0));
	ReturnErrIf(nodeDataPlus(p->nodeRM, 1.0));
	ReturnErrIf(nodeDataPlus(p->nodeSK, 1.0));
	ReturnErrIf(nodeDataPlus(p->nodeSJ, 1.0));
	ReturnErrIf(nodeDataPlus(p->nodeKS, 1.0));
	ReturnErrIf(nodeDataPlus(p->nodeJS, 1.0));
	ReturnErrIf(nodeDataPlus(p->nodeLR, 1.0));
	ReturnErrIf(nodeDataPlus(p->nodeMR, 1.0));
	
	return 0;
}

static int unconfig(device_ *r)
{
	devicePrivate_ *p;
	
	ReturnErrIf(r == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	if(p->rowRName != NULL) {
		free(p->rowRName);
	}
	if(p->rowSName != NULL) {
		free(p->rowSName);
	}
	
	return 0;
}

int deviceConfigTLine(device_ *r, double Z0, double Td)
{
	devicePrivate_ *p;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->private != NULL);
	ReturnErrIf(r->numPins != NUM_PINS);
	ReturnErrIf(Z0 <= 0);
	ReturnErrIf(Td <= 0);
	
	ReturnErrIf((r->pin[K] == &gndRow) && (r->pin[J] == &gndRow));
	ReturnErrIf((r->pin[L] == &gndRow) && (r->pin[M] == &gndRow));
	
	Debug("Configuring %s %p", type, r);
	
	r->type = type;
	r->load = load;
	r->unconfig = unconfig;
	r->step = step;
	r->nextBreak = nextBreak;
	
	r->private =  calloc(1, sizeof(devicePrivate_));
	ReturnErrIf(r->private == NULL);
	p = r->private;
	
	p->Z0 = Z0;
	p->Td = Td;
	p->results = matrixGetResults(r->matrix);
	ReturnErrIf(p->results == NULL);
	
	p->rowRName = malloc(strlen(r->refdes) + 5);
	ReturnErrIf(p->rowRName == NULL);
	strcpy(p->rowRName, r->refdes);
	strcat(p->rowRName, "#in1");
	
	p->rowSName = malloc(strlen(r->refdes) + 5);
	ReturnErrIf(p->rowSName == NULL);
	strcpy(p->rowSName, r->refdes);
	strcat(p->rowSName, "#in2");
	
	p->rowK = r->pin[K];
	ReturnErrIf(p->rowK == NULL);
	p->rowJ = r->pin[J];
	ReturnErrIf(p->rowJ == NULL);
	p->rowR = matrixFindRow(r->matrix, p->rowRName, TYPE_EXTENDED);
	ReturnErrIf(p->rowR == NULL);
	
	p->rowM = r->pin[M];
	ReturnErrIf(p->rowM == NULL);
	p->rowL = r->pin[L];
	ReturnErrIf(p->rowL == NULL);
	p->rowS = matrixFindRow(r->matrix, p->rowSName, TYPE_EXTENDED);
	ReturnErrIf(p->rowS == NULL);
	
	p->nodeRK = matrixFindNode(r->matrix, p->rowR, p->rowK);
	ReturnErrIf(p->nodeRK == NULL);
	p->nodeRJ = matrixFindNode(r->matrix, p->rowR, p->rowJ);
	ReturnErrIf(p->nodeRJ == NULL);
	p->nodeKR = matrixFindNode(r->matrix, p->rowK, p->rowR);
	ReturnErrIf(p->nodeKR == NULL);
	p->nodeJR = matrixFindNode(r->matrix, p->rowJ, p->rowR);
	ReturnErrIf(p->nodeJR == NULL);
	p->nodeRL = matrixFindNode(r->matrix, p->rowR, p->rowL);
	ReturnErrIf(p->nodeRL == NULL);
	p->nodeRM = matrixFindNode(r->matrix, p->rowR, p->rowM);
	ReturnErrIf(p->nodeRM == NULL);
	
	p->nodeSL = matrixFindNode(r->matrix, p->rowS, p->rowL);
	ReturnErrIf(p->nodeSL == NULL);
	p->nodeSM = matrixFindNode(r->matrix, p->rowS, p->rowM);
	ReturnErrIf(p->nodeSM == NULL);
	p->nodeLS = matrixFindNode(r->matrix, p->rowL, p->rowS);
	ReturnErrIf(p->nodeLS == NULL);
	p->nodeMS = matrixFindNode(r->matrix, p->rowM, p->rowS);
	ReturnErrIf(p->nodeMS == NULL);
	p->nodeSK = matrixFindNode(r->matrix, p->rowS, p->rowK);
	ReturnErrIf(p->nodeSK == NULL);
	p->nodeSJ = matrixFindNode(r->matrix, p->rowS, p->rowJ);
	ReturnErrIf(p->nodeSJ == NULL);	
	
	p->nodeRR = matrixFindNode(r->matrix, p->rowR, p->rowR);
	ReturnErrIf(p->nodeRR == NULL);	
	p->nodeSS = matrixFindNode(r->matrix, p->rowS, p->rowS);
	ReturnErrIf(p->nodeSS == NULL);
	p->nodeKS = matrixFindNode(r->matrix, p->rowK, p->rowS);
	ReturnErrIf(p->nodeKS == NULL);	
	p->nodeJS = matrixFindNode(r->matrix, p->rowJ, p->rowS);
	ReturnErrIf(p->nodeJS == NULL);
	p->nodeLR = matrixFindNode(r->matrix, p->rowL, p->rowR);
	ReturnErrIf(p->nodeLR == NULL);	
	p->nodeMR = matrixFindNode(r->matrix, p->rowM, p->rowR);
	ReturnErrIf(p->nodeMR == NULL);	
	
	return 0;
}
